#include <cstdlib>
#include <unistd.h>
#include <dirent.h>
#include <msp/core/application.h>
#include <msp/core/systemerror.h>
#include <msp/strings/regex.h>
#include <msp/strings/utils.h>
#include "dir.h"
#include "path.h"
#include "stat.h"
#include "utils.h"

using namespace std;

namespace Msp {
namespace FS {

namespace
{

enum
{
#ifdef _WIN32
	ITEMSEP = ';'
#else
	ITEMSEP = ':'
#endif
};

/** Helper function to determine the location of the program's executable.
Caches the last result to cut down filesystem access with repeated calls. */
const Path &get_bin_dir(const string &argv0)
{
	static string last_argv0;
	static Path bin_dir;

	if(!(argv0==last_argv0))
	{
		Path exe;
		if(argv0.find('/')==string::npos)
		{
			const char *path = getenv("PATH");
			vector<string> dirs = split(path, ITEMSEP);
			for(vector<string>::const_iterator i=dirs.begin(); i!=dirs.end(); ++i)
				if(exists(Path(*i)/argv0))
				{
					exe = realpath(Path(*i)/argv0);
					break;
				}
		}

		if(exe.empty())
			exe = realpath(argv0);

		last_argv0 = argv0;
		bin_dir = dirname(exe);
	}

	return bin_dir;
}

}


not_a_directory::not_a_directory(const Path &p):
	runtime_error(p.str())
{ }


void mkpath(const Path &path, int mode)
{
	Path p;
	for(Path::Iterator i=path.begin(); i!=path.end(); ++i)
	{
		p /= *i;
#ifdef _WIN32
		if(p.size()==1 && p.is_absolute())
			continue;
#endif
		if(FS::Stat st = stat(p))
		{
			if(!st.is_directory())
				throw not_a_directory(p);
			continue;
		}
		else
			mkdir(p, mode);
	}
}

void rmpath(const Path &path)
{
	list<string> files = list_files(path);
	for(list<string>::iterator i=files.begin(); i!=files.end(); ++i)
	{
		Path p = path / *i;
		if(is_dir(p))
			rmpath(p);
		else
			unlink(p);
	}

	rmdir(path);
}

list<string> list_files(const Path &path)
{
	return list_filtered(path, string());
}

list<string> list_filtered(const Path &path, const string &filter)
{
	Regex r_filter(filter);

	list<string> result;
	DIR *dir = opendir(path.str().c_str());
	if(!dir)
		throw system_error("opendir");

	while(dirent *de = readdir(dir))
	{
		const char *fn = de->d_name;
		if(fn[0]=='.' && (fn[1]==0 || (fn[1]=='.' && fn[2]==0))) 
			continue;
		if(r_filter.match(fn))
			result.push_back(fn);
	}
	closedir(dir);

	return result;
}

Path get_sys_conf_dir(const string &)
{
	return get_sys_conf_dir();
}

Path get_sys_conf_dir()
{
	const char *argv0 = Application::get_argv0();
	if(!argv0)
		throw logic_error("no startup command");

	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
	{
		dir /= "..";
		if(dir[-1]=="usr")
			dir /= "..";
		return dir/"etc";
	}
	else
		return dir;
}

Path get_sys_data_dir(const string &, const string &)
{
	return get_sys_data_dir();
}

Path get_sys_data_dir()
{
	const char *argv0 = Application::get_argv0();
	if(!argv0)
		throw logic_error("no startup command");

	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
		return dir/".."/"share"/Application::get_name();
	else if(dir[-1]=="MacOS")
		return dir/".."/"Resources";
	else
		return dir;
}

Path get_sys_lib_dir(const string &, const string &)
{
	return get_sys_lib_dir();
}

Path get_sys_lib_dir()
{
	const char *argv0 = Application::get_argv0();
	if(!argv0)
		throw logic_error("no startup command");

	Path dir = get_bin_dir(argv0);

	if(dir[-1]=="bin" || dir[-1]=="sbin")
		return dir/".."/"lib"/Application::get_name();
	else
		return dir;
}

Path path_lookup(const string &name, const list<Path> &paths)
{
	for(list<Path>::const_iterator i=paths.begin(); i!=paths.end(); ++i)
	{
		Path full = *i/name;
		if(exists(full))
			return realpath(full);
	}

	return Path();
}

Path path_lookup(const string &name)
{
	const char *path = getenv("PATH");
	vector<string> dirs = split(path, ITEMSEP);
	return path_lookup(name, list<Path>(dirs.begin(), dirs.end()));
}

} // namespace FS
} // namespace Msp
